package pruebasGaleShapley;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;



/**
 * <p>Clase que lee dos archivos de texto con listas de preferencias y ejecuta el algoritmo de GaleShapley
 * para emparejar los miembros de cada uno de ellos.</p>
 * <p>Muestra las preferencias de cada miembro, las parejas estables formadas y una muestra de parejas no estables</p>
 * <p>Adicionalmente, guarda en un archivo parejas.txt las parejas estables formadas</p>
 * @author Alberto Hueso Alonso
 *@since 07/12/2017
 *@see Funciones#emparejarGaleShapleyCompleto(Map, Map, ArrayList)
 *@see Funciones#leerElementoPreferencias(String)
 *@see Funciones#verParejas(Map, String)
 *@see Funciones#verParejas(ArrayList, String) 
 *@see #storeParejas(Map, String)
 * */
@SuppressWarnings("rawtypes")
public class GaleShapleyUnitario {
	 static ArrayList  noEmparejados=new ArrayList();
	
	static  Map  emparejados=new HashMap();
	static Map<String,ArrayList> elementosX=null;
	static Map<String, ArrayList> elementosY=null;
	
	/**
	 * Acción de la clase GaleShapley
	 * lee dos archivos de texto con listas de preferencias y ejecuta el algoritmo de GaleShapley
	 * para emparejar los miembros de cada uno de ellos
	 * Muestra las preferencias de cada miembro, las parejas estables formadas y una muestra de parejas no estables
	 * Adicionalmente, guarda en un archivo parejas.txt las parejas estables formadas
	 * @param args Nombres de los archivos que guardan las preferencias de los dos grupos
	 */
	public static void main(String[] args) {
			String xFile=null;		//Nombre del archivo que guarda los archivos X
			String yFile=null;		//Nombre del archivo que guarda los archivos Y
			
			//Comprobamos si hay argumentos y si los hay se asignan como los nombres del archivo	
			try{	
			 xFile=args[0];
			 yFile=args[1];
			}catch (IndexOutOfBoundsException e){
				System.out.println("No se han encontrado argumentos\nSe intentarán leer los archivos por defecto");
			}
				
			//Leemos los archivos
			try {
				if (xFile!=null&&yFile!=null){
					elementosX=Funciones.leerElementoPreferencias(xFile);
					elementosY=Funciones.leerElementoPreferencias(yFile);
					System.out.println("Leídos los argumentos: "+xFile+" y "+yFile);
				}else
				{
				elementosX=Funciones.leerElementoPreferencias("elementosX.out");
				elementosY=Funciones.leerElementoPreferencias("elementosY.out");
				System.out.println("Leídos los archivos por defecto: elementosX.out y elementosY.out");
				}
			} catch (IOException e) 
			{
			
			System.err.println("Ha habido algún problema en la lectura de los archivos\n"+e.getMessage());
			}
			
			if (elementosX!=null&&elementosY!=null){
			//Los emparejamos	 
			 emparejados=Funciones.emparejarGaleShapleyCompleto( elementosX, elementosY, noEmparejados);
			
			 //Guardamos las parejas estables en un archivo de texto para utilizarlo posteriormente	 
			 storeParejas(emparejados,"parejas.txt");
			 
			//Mostramos sus preferencias
			Funciones.verPreferencias(elementosX,elementosY);
			
			//Mostramos las parejas estables
			Funciones.verParejas(emparejados, "PAREJAS ESTABLES:\n");
			
			//Mostramos las parejas rotas durante el proceso
			Funciones.verParejas(noEmparejados, "PAREJAS NO ESTABLES:\n");
						
			}
	}	 
	
	
	
	/**
	 * Guarda un archivo de texto con el nombre indicado por nombreFichero con el mapa
	 * <p>Este archivo contendrá en cada línea el valor del mapa, seguido de ":" y la clave</p>
	 * @param mapa Map que se utilizará
	 * @param nombreFichero String nombre del fichero que se guardará
	 */
	// Basada en ejemplo de https://blogdeaitor.wordpress.com/2008/11/21/java-escribir-en-fichero/
	private static void storeParejas(Map mapa, String nombreFichero){


		//Creamos nuestro String que vamos a escribir en el fichero
		String texto="";

		//Creamos el nombre de nuestro fichero  

		try{

			//Creamos un Nuevo objeto FileWriter dandole
			//como parámetros la ruta y nombre del fichero
			FileWriter fichero = new FileWriter(nombreFichero);

			Iterator it = mapa.keySet().iterator();

			while (it.hasNext()){
				String agente = (String) it.next();
				String pareja =(String) mapa.get(agente);
				texto=texto.concat(pareja+":");
				texto=texto.concat(agente+"\r\n");
			}  				

			fichero.write(texto);
			
			//cerramos el fichero
			fichero.close();
			System.out.println("Fichero con las parejas estables creado: "+nombreFichero);
		}catch(Exception ex){
			ex.printStackTrace();
		}
	}
	}


