package tfgMarriageStable;


import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;

/**
 * <p>Clase que se registra como agente que ofrece el servicio "clasificador", tipo "clasificacion"</p>
 * <p>Escucha mensajes con conversationId="archivoDatos"</p>
 * <p>Interpreta el contenido de los mensajes recibidos como un nombre de archivo para trabajar con él</p>
 * <p>Cuando recibe un mensaje ejecuta el método <i>UtilizarDatos</i> que no tiene contenido, se sobreescribirá en las clases hijas
 * @author Alberto Hueso Alonso
 *@since 12-12-2017
 *@see #UtilizarDatos()
 */
public abstract class Clasificador extends Agent {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = -6647249614605440161L;
	/**
	 * Ruta del archivo que se estudiará
	 */
	protected String nameFile;
	/**
	 * El propio agente
	 */
	private Agent esteAgente=this;
	
	
	/**
	 * <p>Recupera el primer argumento introducido y actualiza el campo <i>nameFile</i> con ese valor
	 *<p>Registra el clasificador y el lenguaje SL</p>
	 *<p>Abre el escuchador <i>EscucharNombreArchivo</i>
	 *@see #RegistrarClasificador()
	 *@see Funciones#registrarLenguajeSL
	 *@see EscucharNombreArchivo
	 */
	public void setup(){
		System.out.println("Se ha creado el agente "+esteAgente.getLocalName());
		
		try {
		//Obtenemos, si lo hay, el nombre del archivo introducido como argumento
		nameFile=(String) getArguments()[0];
		}
		catch (Exception e){
			System.out.println("No se ha introducido ningún argumento por consola para el clasificador");
			
			//Registramos el agente
			RegistrarClasificador();
			
			//Registramos el lenguaje SL en el agente
		   	Funciones.registrarLenguajeSL(this);  
		   	
		   	addBehaviour(new EscucharNombreArchivo("archivoDatos"));
		}
		
	   	
	   	
	}
	/**
	 * Muestra un mensaje al eliminarse el agente
	 */
	public void takeDown(){
		System.out.println(esteAgente.getLocalName()+" se ha desregistrado");
	}
	
/**
 * <p>Registra el clasificador en en el Domain Facilitator</p>
 * <p>El nombre del servicio será "clasificador", el tipo "clasificacion" y el propietario "TFG"</p>
 */
private void RegistrarClasificador(){
    	
    	
    	//Petición de registro en el DF
    	DFAgentDescription dfd=new DFAgentDescription();
    	
    	//Servicio que se proporciona
    	ServiceDescription sd= new ServiceDescription();  
    	
    	//Asignamos el nombre al servicio
    	sd.setName("clasificador");
    	
    	
    	//Se añade el tipo y el propietario
    	sd.setType("clasificacion");
    	sd.setOwnership("TFG");
    	
    	
    	//Añadimos los servicios al descriptor de agentes
    	dfd.addServices(sd);
    	dfd.setName(getAID());
    	
    	try {
    		//Añadimos al registro
			DFService.register(this, dfd);
			System.out.println("Se ha registrado el agente "+getLocalName());
		} catch (FIPAException e) {				
			
				System.err.println("No se ha podido registrar el agente "+getLocalName());
				
				
						
		}
    	
    }


/**
 * <p>Comportamiento que recibe mensajes con un conversationId que se determina en el constructor</p>
 * <p>Cuando recibe un mensaje así actualiza el campo <i>nameFile</i> con el contenido del mismo</p>
 * <p>Por último, ejecuta el método <i>UtilizarDatos()</i>
 *@see Clasificador#UtilizarDatos()
 */
private class EscucharNombreArchivo extends Behaviour{
	
	
	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = -6393544394957670474L;
	/**
	 * Identificador de conversación en el que escucha el comportamiento
	 */
	private String conversationId;
	
	/**
	 * Constructor 
	 * @param pConversation  String que representa el conversationId en el que se escuchará
	 */
	private EscucharNombreArchivo(String pConversation){
		this.conversationId=pConversation;
	}
	@Override
	public void action() {
		//Creamos un filtro, solo se recibirán los mensajes cuyo identificador de conversacion coincida con conversationId
		
		MessageTemplate mt=MessageTemplate.MatchConversationId(conversationId);
		//Recepción no bloqueante de un mensaje según el filtro creado mt
		ACLMessage msg = myAgent.receive(mt);			
				
	if (msg != null) {
		
		
		System.out.println(esteAgente.getLocalName()+": Mensaje recibido de "+msg.getSender().getLocalName()+"= "+msg.getContent());
		nameFile=msg.getContent();
		
		//Añadimos la acción por defecto al recibir el mensaje
		UtilizarDatos();
					
	}else
		{
		//Bloqueamos el comportamiento hasta recibir un mensaje en la cola
		block();
		}
  }

			@Override
			public boolean done() {
				// TODO Auto-generated method stub
				return false;
			}
}

/**
 * <p>No ejecuta ninguna acción, los detalles se implementarán en las clases hijas</p>
 */
protected void UtilizarDatos(){
	//Se escribirá la acción en las subclases instanciables
	}
	


}