package tfgMarriageStable;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import jade.content.lang.Codec;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.basic.Action;
import jade.core.Agent;
import jade.domain.JADEAgentManagement.JADEManagementOntology;
import jade.domain.JADEAgentManagement.ShutdownPlatform;
import jade.lang.acl.ACLMessage;
import jade.util.leap.ArrayList;
import jade.util.leap.HashMap;
import jade.util.leap.Iterator;
import jade.util.leap.Map;
import jade.util.leap.Set;

/**
 * Clase que guarda funciones auxiliares
 * @author Alberto Hueso Alonso
 *@since 06/12/2017
 */
public  class Funciones {
	/**
	 * Obtiene un número aleatorio entre 0 y 1 con dos decimales
	 * @return un número con dos decimales entre 0 y 1 
	 */
	public static float getDecimal(){
		return  (float) Math.round(Math.random()*100)/100;
	}
		/**
	 * Obtiene un número aleatorio entre 0 y 10 con un decimal
	 * @return int Un número con  entre 0 y 10
	 */
	public static float getMultiplicador(){
		float numero=Math.round(Math.random()*100);
		return  numero/10;
		
		
	}
	

	/**
	 * Función que registra el lenguaje SL en el agente agent
	 * @param pAgent Agent que registrará el lenguaje
	 */
	public static void registrarLenguajeSL(Agent pAgent){
		try
    	{
    	//Instanciamos un codec SL
    	Codec codec = new SLCodec();
    	//Recuperamos el nombre estándard del lenguaje
    	String codec_name = codec.getName();
    	//Registrem el llenguatge SL
    	pAgent.getContentManager().registerLanguage( codec, codec_name );
    	 	
    		    	}
    	catch( Exception e ){
    	System.err.println( "Error registrando el lenguaje" );}
	}
	
	/**
	 * Función que dado un agente agent, cierra la plataforma
	 * 
	 * @param agent Agent   Agente que solicita el cierre de la plataforma
	 * @see <a href=https://stackoverflow.com/questions/7025683/stop-a-jade-system-java-agents>Adaptada desde</a>
	 */
	//Adaptada desde https://stackoverflow.com/questions/7025683/stop-a-jade-system-java-agents
	public static void cerrarSistema(Agent agent){
		Codec codec = new SLCodec();    
		Ontology jmo = JADEManagementOntology.getInstance();
		agent.getContentManager().registerLanguage(codec);
		agent.getContentManager().registerOntology(jmo);
		ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
		msg.addReceiver(agent.getAMS());
		msg.setLanguage(codec.getName());
		msg.setOntology(jmo.getName());
		try {
		    agent.getContentManager().fillContent(msg, new Action(agent.getAID(), new ShutdownPlatform()));
		    agent.send(msg);
		}
		catch (Exception e) {}
	}  
	
	/**
	 * <p>Empareja los elementos de dos grupos dadas sus preferencias según el algoritmo de Gale-Shapley</p>
	 * <p>A medida que forma las parejas añade <strong>solo las parejas provisionales que rompen</strong> a noEmparejados</p>
	 * @param  elementosX Map con los elementos y preferencias de una clase
	 * @param elementosY Mapcon los elementos y preferencias de la otra clase
	 * @param noEmparejados ArrayList con los elementos no emparejados
	 * @return Mapa con parejas estables. Las claves serán los nombres de elementosY. Los valores los de X
	 */	
@SuppressWarnings("unchecked")
public static Map emparejarGaleShapley(Map elementosX, Map elementosY, ArrayList noEmparejados){
		
	//Iniciamos el contador de tiempo		  
	  long time_start = System.currentTimeMillis();
	  //Creamos un nuevo mapa para las parejas formadas
	  Map parejasFormadas = new HashMap();
	  
	  	//Elementos xLibresList representa los elementosX todavía no emparejados, inicialmente son todos
		ArrayList xLibresList = new ArrayList();
		
		//Recorremos los elementosX y los añadimos a la lista de elementos libres X
		Set xSet= elementosX.keySet(); //Set con las claves de elementosX, es decir, todos los localname de los elementos X
		java.util.Iterator<String> it = xSet.iterator();
		while (it.hasNext()){
			xLibresList.add(it.next());
			
		}
		
		//Mientras la lista de elementos no está libre la recorremos e intentamos emparejar a cada X con su favorito Y
		while (!xLibresList.isEmpty()){
			System.out.println("emparejando");
			
				//Eliminamos de la lista de libres
				String currentX= (String) xLibresList.remove(0);
				 //Obtenemos la lista de parejas del elemento completo
			   ArrayList parejasPreferidasActual = (ArrayList) elementosX.get(currentX);
			    
			    
			    
			    Iterator i=parejasPreferidasActual.iterator();
			    String y;//posible pareja de currentX
			    
			    //Recorremos la lista de parejas de currentX
			    while(i.hasNext()){
			    	y=(String) i.next();//pareja de CurrentX en cada momento
			    	if(parejasFormadas.get(y) == null) { // Este elemtentoY no está emparejado
						//Se emparejan los dos
		                parejasFormadas.put(y, currentX);
		                break;
		                
		                
		               /*Si el elementoY sí tiene pareja 
		                se comprueba si esta pareja está antes en su lista de preferencias que currentX o no
		                Si la pareja actual está antes se mantiene esa pareja.
		                Si, en cambio, y prefiere a currentX cambia su pareja por él
		                y la pareja anterior se añade a la lista de elementos libres
		               */
		            }else {
		                String otherX = (String) parejasFormadas.get(y);
		                ArrayList parejasDeY = ((ArrayList) elementosY.get(y));
		               
		                if(parejasDeY.indexOf(currentX) < parejasDeY.indexOf(otherX)){
		                    //Este y prefiere a currentX antes que con quien está
		                    parejasFormadas.put(y, currentX);//Se forma la pareja
		                    Pareja noPareja=new Pareja(y,otherX);
		                    
		                  //Se añade a la lista de parejas no estables
		                    if (!checkListaContainsPareja(noEmparejados,noPareja))
		                    noEmparejados.add(noPareja);
		                    
		                    xLibresList.add(otherX);//Se añade otherX a la lista de elementos libres
		                    break;
		                }
		            }
			   
			    }
			    
		}
		
		//Finalizamos el contador de tiempo
  	 long time_end = System.currentTimeMillis();
  	 
  	 //Imprimimos el tiempo empleado en el emparejamiento
  	 System.out.println("\n\nEl emparejado ha tomado "+ ( time_end - time_start ) +" millisegundos");
		return parejasFormadas;
			
			
		
	}

/**
 * <p>Empareja los elementos de dos grupos dadas sus preferencias según el algoritmo de Gale-Shapley</p>
 * <p>A medida que forma las parejas añade <strong>todas las parejas que se rompen y las que no se llegan a formar porque los elementos prefieren a sus parejas provisionales antes que a un nuevo pretendiente</strong> a noEmparejados</p>
 * @param  elementosX Map con los elementos y preferencias de una clase
 * @param elementosY Mapcon los elementos y preferencias de la otra clase
 * @param noEmparejados ArrayList con los elementos no emparejados
 * @return Mapa con parejas estables. Las claves serán los nombres de elementosY. Los valores los de X
 */	
@SuppressWarnings("unchecked")
public static Map emparejarGaleShapleyCompleto(Map elementosX, Map elementosY, ArrayList noEmparejados){
	
//Iniciamos el contador de tiempo		  
  long time_start = System.currentTimeMillis();
  //Creamos un nuevo mapa para las parejas formadas
  Map parejasFormadas = new HashMap();
  
  	//Elementos xLibresList representa los elementosX todavía no emparejados, inicialmente son todos
	ArrayList xLibresList = new ArrayList();
	
	//Recorremos los elementosX y los añadimos a la lista de elementos libres X
	Set xSet= elementosX.keySet(); //Set con las claves de elementosX, es decir, todos los localname de los elementos X
	java.util.Iterator<String> it = xSet.iterator();
	while (it.hasNext()){
		xLibresList.add(it.next());
		
	}
	
	//Mientras la lista de elementos no está libre la recorremos e intentamos emparejar a cada X con su favorito Y
	while (!xLibresList.isEmpty()){
		System.out.println("emparejando");
			//Eliminamos de la lista de libres
			String currentX= (String) xLibresList.remove(0);
			 //Obtenemos la lista de parejas del elemento completo
		   ArrayList parejasPreferidasActual = (ArrayList) elementosX.get(currentX);
		    
		   
		    
		    
		    Iterator i=parejasPreferidasActual.iterator();
		    String y;//posible pareja de currentX
		    
		    //Recorremos la lista de parejas de currentX
		    while(i.hasNext()){
		    	y=(String) i.next();//pareja de CurrentX en cada momento
		    	if(parejasFormadas.get(y) == null) { // Este elementoY no está emparejado
					//Se emparejan los dos
	                parejasFormadas.put(y, currentX);
	                break;
	               /*Si el elementoY sí tiene pareja 
	                se comprueba si esta pareja está antes en su lista de preferencias que currentX o no
	                Si la pareja actual está antes se mantiene esa pareja.
	                Si en cambio y prefiere a currentX cambia su pareja por él
	                y la pareja anterior se añade a la lista de elementos libres
	               */
	            }else {
	                String otherX = (String) parejasFormadas.get(y);
	                ArrayList parejasDeY = ((ArrayList) elementosY.get(y));
	               
	                
	                if(parejasDeY.indexOf(currentX) < parejasDeY.indexOf(otherX)){
	                    //Este y prefiere a currentX antes que con quien está
	                    parejasFormadas.put(y, currentX);//Se forma la pareja
	                    Pareja noPareja=new Pareja(y,otherX);
	                    if (!checkListaContainsPareja(noEmparejados,noPareja)){
	                    noEmparejados.add(noPareja);
	                    }
	                    xLibresList.add(otherX);//Se añade otherX a la lista de elementos libres
	                    
	                    break;
	                }else {
	                	Pareja noPareja=new Pareja(y,currentX);
	                	if (!checkListaContainsPareja(noEmparejados,noPareja)){
	                    noEmparejados.add(noPareja);//Añadimos a la lista de no emparejados la pareja que no se ha llegado a formar
	                	}
	                	}
	            }
		   
		    }
		    
	}
	
	//Finalizamos el contador de tiempo
	 long time_end = System.currentTimeMillis();
	 
	 long time=time_end-time_start;
	 
	 //Imprimimos el tiempo empleado en el emparejamiento
	 System.out.println("\n\nEl emparejado ha tomado "+ time +" milisegundos");
	 
	 return parejasFormadas;
		
		
	
}
/**
 * Muestra las parejas de un Mapa con un pTexto inicialmente
 * @param emparejados Map con un grupo de parejas
 * @param pTexto     String texto que se mostrará como inicio
 */
public static void verParejas(Map emparejados,String pTexto) {
	//Control emparejados
		System.out.println(pTexto);
			Iterator i=emparejados.keySet().iterator();
			
			while (i.hasNext()){
				String elegido = (String) i.next();
				String elector=(String) emparejados.get(elegido);
				System.out.println(elector+","+elegido);
			}
		
			System.out.println("\n\n PAREJAS: "+emparejados.size());
			System.out.println("\n\n");
	
}
/**
 * Muestra las parejas de un ArrayList y un texto inicial pTexto
 * @param emparejados  ArrayList con un grupo de parejas
 * @param pTexto       String texto que se mostrará como inicio
 */
public static void verParejas(ArrayList emparejados,String pTexto) {
	//Control emparejados
		System.out.println(pTexto);
			Iterator i=emparejados.iterator();
			
			while (i.hasNext()){
				Pareja pareja = (Pareja) i.next();
				String x=pareja.getX();
				String y=pareja.getY();
				System.out.println(x+","+y);
			}
		
			System.out.println("\n\n PAREJAS: "+emparejados.size());
			System.out.println("\n\n");
	
}
/**
 * <p>Función que dados dos mapas de elementos con las listas de preferencias de parejas de cada uno muestra en pantalla las parejas preferidas de cada elemento de mayor a menor</p>
 * <p>En primer lugar la clave	y luego el valor: la lista de preferencias</p>
 * @param elementosX Mapa con los elementos  y sus preferencias de una clase </p>	
 * @param elementosY Mapa con los elementos y sus preferencias de la otra clase
 */
public static void  verPreferencias(Map elementosX,Map elementosY){
  //Control
 System.out.println("\n\n...........................................\n\n   LISTAS DE PREFERENCIAS DE LAS PAREJAS\n...........................................\n");
	Iterator i = elementosX.keySet().iterator();
	while (i.hasNext()){
		String agente = (String) i.next();
		
		ArrayList a=(ArrayList) elementosX.get(agente);
		
		
		System.out.println(agente+": "+a);						
	}
	//Control
	System.out.println("\n........................\n");
	Iterator i3 = elementosY.keySet().iterator();
	while (i3.hasNext()){
		String agente = (String) i3.next();
		
		ArrayList a=(ArrayList) elementosY.get(agente);
		
		
		System.out.println(agente+": "+a);						
	}
	System.out.println("\n\n");
}
/**
 * Une dos mapas del mismo tipo, añade al mapa objetivo los elementos de mapaParaUnir
 * @param objetivo Map al que se unirá el segundo parámetro 
 * @param mapaParaUnir Map que se unirá al primer parámetro
 * @return Map con los elementos de ambos mapas unidos
 */
public static Map joinMaps(Map objetivo,Map mapaParaUnir){
	
	Iterator it = mapaParaUnir.keySet().iterator();
	while (it.hasNext()){
		Object clave = it.next();
		Object objeto=mapaParaUnir.get(clave);
		objetivo.put(clave, objeto);
	}
	return objetivo;
	
}

/**
 * Une dos ArrayList del mismo tipo 
 * @param objetivo ArrayList al que se unirán los elementos
 * @param listaParaUnir ArrayList que se unirá a objetivo
 * @return ArrayList con los miembros de ambas listas unidas.
 */
public static ArrayList joinArrays (ArrayList objetivo, ArrayList listaParaUnir){
	
	Iterator it=listaParaUnir.iterator();
	while (it.hasNext()){
		objetivo.add(it.next());
	}
	return objetivo;
}

/**
 * Obtiene un String con la fecha y hora en el formato yyyy_MM_dd_HH_mm_ss
 * @return String con fecha y hora
 */
public static String getDate(){
	 DateFormat sdf = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss");
	Date date=new Date();
	String fecha=sdf.format(date);
	return fecha;
	
}
/**
 * Escribe un archivo de texto con el nombre nameFile y el texto pTexto
 * @param pTexto String que se incluirá en el archivo
 * @param nameFile String nombre del archivo
 * @param continuarEnFichero boolean indica si se continuará escribiendo en el archivo (true) o si se reescribirá únicamente con pTexto
 */
public static void createFile(String pTexto, String nameFile, Boolean continuarEnFichero){
	//Creamos un Nuevo objeto FileWriter dandole
    //como parámetros la ruta con el nombre del fichero y si se reescribirá o continuará en el archivo existente
    FileWriter fichero;
	try {
		fichero = new FileWriter(nameFile,continuarEnFichero);
	
    //Creamos un buffer para él
    BufferedWriter bw = new BufferedWriter(fichero);
    //Creamos un printwriter para ese buffer
	  PrintWriter out = new PrintWriter(bw);
	 out.print(pTexto);
	  
    
	 //Cerramos 
	 out.close();
	 bw.close();		   
    fichero.close();
	  }
    catch (IOException e) {
			e.printStackTrace();
		}
	
}
/**
 * Función que comprueba si una lista contiene una pareja
 * @param lista
 * @param pareja
 * @return Boolean , true si la contiene y false en caso contrario
 */
private static boolean checkListaContainsPareja(ArrayList lista,Pareja pareja){
	boolean resultado = false;
	Iterator it=lista.iterator();
	while (it.hasNext()){
		Pareja p=(Pareja) it.next();
		
		if (p.getX().equals(pareja.getX())&&p.getY().equals(pareja.getY())){
			
			return true;
		}
			
	}
	return resultado;
	
}
}
