package tfgMarriageStable;


import java.io.IOException;
import java.util.Iterator;
import javax.swing.JOptionPane;
import jade.content.lang.sl.SLCodec;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.core.behaviours.SequentialBehaviour;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPANames;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.FailureException;
import jade.domain.FIPAAgentManagement.NotUnderstoodException;
import jade.domain.FIPAAgentManagement.RefuseException;
import jade.domain.FIPAAgentManagement.SearchConstraints;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.lang.acl.UnreadableException;
import jade.proto.AchieveREResponder;
import jade.util.leap.ArrayList;
import jade.wrapper.AgentController;
import jade.wrapper.ContainerController;
import jade.wrapper.StaleProxyException;

/**
 * <p>Agente que gestionará el resto de agentes</p>
 * <p>Recibe los datos del usuario y envía los argumentos necesarios a cada agente</p>
 * <p>Crea un Emparejador (GaleShapley o GaleShapleyCompleto según indique el usuario) 
 * y un agente Clasificador Weka</p>
 * <p>Crea tantos pares de ElementoX y elementoY como número de parejas haya indicado el usuario</p>
* <p>Una vez que ha creado todos los agentes, busca los elementos de los grupos X e Y. Si no es capaz de encontrarlos tras 50 intentos obteniendo siempre el mismo resultado insuficiente muestra un mensaje de error y cierra el programa.</p>
 * <p>Si la búsqueda es satisfactoria, les envía la lista de elementos del otro grupo a cada elemento para que puedan preguntarles por sus características</p> 
 * <p>Abre canales para escuchar a otros agentes, especialmente a los emparejadores que le pedirán realizar
 * nuevas iteraciones, esto es, eliminar todos los elementos X e Y y crear nuevas parejas</p>
 *<p>Cuando se han realizado todas las iteraciones responde a los emparejadores con Agree, en caso contrario con Refuse</p>
 * 
 * @author Alberto Hueso Alonso
 * @since 10/12/2017
 * @see Emparejador
 * @see GaleShapley
 * @see GaleShapleyCompleto
 * @see Elemento
 * @see ElementoX
 * @see ElementoY
 * @see Weka
 */
public class GestorAgentes extends Agent{
	/**
	 * Constante para serializar
	 */
	private static final long serialVersionUID = 6822051805711997697L;
	/**
	 *Número máximo de intentos que buscará parejas, encontrando el mismo número de parejas sin llegar al esperado
	 */
	static final int NUMERO_INTENTOS_BUSQUEDAS=50;
	/**
	 * Número máximo de resultados en las búsquedas por defecto. En la versión de JADE utilizada, 4.5, es de 100
	 */
	static final int MAXIMO_RESULTADO_BUSQUEDAS_DEFECTO=100;

	/**
	 * Número máximo de parejas permitidas
	 */
	static final int PAREJAS_MAXIMAS_PERMITIDAS =400 ;
	
	/**
	 * Número máximo de características permitidas
	 */
	static final int NUMERO_MAXIMO_CARACTERISTICAS=10;
	
	/**
	 * Número máximo del producto del número de parejas y el número de iteraciones
	 */
	static final int NUMERO_MAXIMO_PAREJASxITERACIONES=2400;
	//Variables
	/**
	 * El propio agente
	 */
	private Agent esteAgente=this;
	/**
	 * Guarda el número de parejas creadas
	 */
	
	private int contadorParejas=0;
	/**
	 * Guarda el número de parejas que ha de crear
	 */
	private int numeroParejas=0;
	
	/**
	 * Guarda el número que se asignará al agente que se está creando
	 */
	private int numeroAgenteNombre=0;
	/**
	 * Guarda si se ha de cerrar el sistema
	 */
	private boolean shutdown=false;
	
	/**
	 * Guarda el número de características
	 */
	private int numeroCaracteristicas=0;
	/**
	 * Guarda si hay que guardar los archivos para pruebas
	 */
	private boolean filePruebas=false; //Guarda si hay que crear el archivo para muestras
	/**Guarda si hay que guardar los resultados en un nuevo archivo(true) o sobreescribir el archivo por defecto(false)*/
	private boolean newfile=false;
	
	/**
	 *Guardará la lista de controladores de los agentes creados
	 */
	private ArrayList listaControladoresAgentes=new ArrayList();

	/**
	 * Guarda el número de iteraciones que hay que realizar
	 */
	private int numeroIteraciones=1; 
	/**
	 * Guarda la iteración que se está realizando en un momento dado
	 */
	private int iteracionActual=1;//Iteración actual
	
	/**
	 * Guarda si se creará un emparejador GaleShapley(false) o GaleShapleyCompleto(true)
	 */
	private boolean GaleShapleyCompleto=false;
	
	/**
	 * Si no se encuentran todos los agentes que se buscan
	 */
	private boolean faltanAgentes=false;
	
	/**
	 * Guarda la lista de parejas de elementos X
	 */
	private ListaParejas listaParejasX=new ListaParejas();
	
	/**
	 * Guarda la lista de parejas de elementos Y
	 */
	private ListaParejas listaParejasY=new ListaParejas();
	/**
	 * Número de agentes encontrados en la búsqueda
	 */
	private int agentesEncontrados=0;
	/**
	 * Guarda el número de veces que se ha encontrado el límite designado de búsqueda
	 */
	private int contadorEncontradoLimite=0;
	
	/**
	 * Número emparejadores creados
	 */
	private int numeroEmparejadores=0;
	
	/**
	 * Guarda el valor que se pasará a los elementos para que modulen sus preferencias o no
	 */
	private boolean moduladas=false;
	
	
	
	
	
	/**
	 * <p>Inicialización del agente </p>
	 * <p>Recupera los datos introducidos por el usuario</p>
	 * <p>Si el producto del número de parejas por las iteraciones supera NUMERO_MAXIMO_PAREJASxITERACIONES se cierra la plataforma</p>
	 * <p>Crea el emparejador, GaleShapley o GaleShapleyCompleto según los argumentos introducidos</p>
	 * <p>Crea los elementos X Y</p>
	 * <p>Busca los elementos de cada grupo y los introduce en sendas listas</p>
	 * <p>Envía a cada elemento la lista de elementos del otro grupo</p>
	 * <p>Abre los escuchadores y manejadores de Requests</p>
	 *@see #getUserData()
	 *@see GaleShapley
	 * @see GaleShapleyCompleto
	 * @see  crearElementosXY
	 * @see BusquedaDeElementos
	 * @see sendListasParejas
	 * @see ManejadorResponderRequests
	 */
	public void setup(){		
		
		//Obtenemos los introducidos por el usuario
		getUserData();
		
		
		//Se comprueba el producto de numeroParejas*numeroIteraciones, si supera NUMERO_MAXIMO_PAREJASxITERACIONES
		if (numeroParejas*numeroIteraciones>NUMERO_MAXIMO_PAREJASxITERACIONES){
			System.err.println("\n\n\t\t\t\t\t||||||ATENCIÓN||||||\n\t\t\tEl número de parejas*el número de iteraciones no puede ser mayor de "+NUMERO_MAXIMO_PAREJASxITERACIONES+
					"\n\t\t\tSe cerrará el programa\n\n\n");
			
			Funciones.cerrarSistema(esteAgente);//Cerramos el sistema
			
		//El programa continua en caso contrario
		}else{
			
			//Si se han recibido el número de parejas y número características se crean el resto de agentes
			if (numeroParejas>0&&numeroCaracteristicas>0){
				
			
				//Creamos el emparejador, según el valor de GaleShapleyCompleto	       
			    if (GaleShapleyCompleto){
			    	crearEmparejador("GaleShapleyCompleto","tfgMarriageStable.GaleShapleyCompleto");
			    
			    }else{
			    	//Creamos el agente emparejador de la subclase tfgMarriageStable.GaleShapley
				    crearEmparejador("GaleShapley","tfgMarriageStable.GaleShapley");	
				
			    }
				
			    //Creamos el agente Weka
			   crearGestorWeka("weka","tfgMarriageStable.Weka");
			   
			
			 //Registramos el lenguaje SL en el agente
			   	Funciones.registrarLenguajeSL(this);  
			   
				//Creamos una plantilla para Requests
				MessageTemplate plantilla = crearPlantillaRequest();
				
			    //Añadimos el manejador de respuesta a los mensajes recibidos según esa plantilla
				addBehaviour(new ManejadorResponderRequests(this, plantilla));
				
				 SequentialBehaviour seq = new SequentialBehaviour();
				 	seq.addSubBehaviour(new crearElementosXY());
					seq.addSubBehaviour(new BusquedaDeElementos());
					seq.addSubBehaviour(new sendListasParejas());
					
				   	//Añadimos el comportamiento secuencial
							addBehaviour(seq);
				
				//Añadimos el comportamiento para escuchar "tiempoInicioConsultas"
				addBehaviour(new EscucharConversationId("tiempoInicioConsultas"));

							
				}
		}
	}
	/**
	 * <p>Muestra mensaje de despedida y cierra el sistema</p>
	 * <p>Si faltanAgentes==true (no se han encontrado los buscados) muestra un mensaje de error</p>
	 * @see Funciones#cerrarSistema(Agent)
	 */
	public void takeDown(){
		System.out.println("El gestor de agentes: "+this.getLocalName()+" se ha borrado, se cerrará el sistema");
		GestorAgentes a=(GestorAgentes) esteAgente;
		
		if (a.faltanAgentes){

	    	//Si faltan Agentes muestra un mensaje de error
	    	
				System.err.println("\n"+this.getLocalName()+"solo ha encontrado"+agentesEncontrados+ " agentes y se esperaban "+numeroParejas*2+"\nOpción1:Intente con un número de parejas menor de "+MAXIMO_RESULTADO_BUSQUEDAS_DEFECTO+"\nOpción2:Asegúrese de que ha aumentado suficientemente el número de búsquedas máximas\nEjemplo de parámetros de ejecución:-jade_domain_df_maxresult 800 gestor:tfgMarriageStable.GestorAgentes\nLa plataforma se cerrará, empiece de nuevo\n\n\n");
					
		}
		Funciones.cerrarSistema(esteAgente);
	}
	
	
	

	
	/**
	 * Crea un agente de la clase pClase (Se espera una subclase de Emparejador)
	 * Le pasa como argumento el número de parejas
	 */
	/**
	 * <p> Crea un agente de la clase pClase (Se espera una subclase de Emparejador)</p>
	 * <p>Pasa como argumento al agente creado el número de parejas, el AID del gestor de agentes y el valor de filePruebas</p>
	 * <ul>
	 * 				<li>args[0]=numeroParejas;</li>
					<li>args[1]=esteAgente.getAID();</li>
					<li>args[2]=filePruebas;		</li>			
					<li>args[3]=newfile;</li>
	 * </ul>
	 * @param pNombre String nombre del agente que se crea
	 * @param pClase  String Nombre de la clase de agente que se crea
	 */
	private void crearEmparejador(String pNombre,String pClase)  {
		ContainerController cc = getContainerController();
	       AgentController ac;
		try {
					//Nombre que se pondrá al agente
					String name=pNombre;
					//Clase que se asignara
					String clase=pClase;
					/*Asignamos los argumentos que se pasarán createNewAgent) y a través de él a los agentes creados
					 * El número de parejas
					 * El AID del managerAgents
					 * Si se ha de crear el archivo de pruebas
					 * Si se ha de guardar un nuevo archivo
					 * Si las preferencias de los agentes son moduladas
					 * 
					 */
					Object[] args=new Object[5];					
					args[0]=this.numeroParejas;
					args[1]=esteAgente.getAID();
					args[2]=filePruebas;					
					args[3]=newfile;
					args[4]=moduladas;		
					
					
					//Creamos el agente y lo iniciamos
					ac = cc.createNewAgent(name, clase, args);
					ac.start(); 
				    numeroEmparejadores++;
					
					
					System.out.println("Crear emparejador finalizado");
		} 
		
		//Excepción que se produce si no se ha podido crear el agente
		catch (StaleProxyException e) {			
			
			System.out.println("No se ha podido crear el emparejador\n"+e.getMessage());
		}
	
	}
	
	private void crearGestorWeka(String pNombre,String pClase)  {
		ContainerController cc = getContainerController();
	       AgentController ac;
		try {
					//Nombre que se pondrá al agente
					String name=pNombre;
					//Clase que se asignara
					String clase=pClase;
					
					
					
					//Creamos el agente y lo iniciamos
					ac = cc.createNewAgent(name, clase,null);
					ac.start(); 
				
					
					
					System.out.println("Crear Weka finalizado");
		} 
		
		//Excepción que se produce si no se ha podido crear el agente
		catch (StaleProxyException e) {			
			
			System.err.println("No se ha podido crear el gestorWeka\n"+e.getMessage());
		}
	
	}
	/**
	 * Crea una plantilla MessageTemplate para protocolo FIPA_REQUEST y Perfomative REQUEST.
	 * @return MessageTemplate La plantilla creada
	 */
	private  MessageTemplate crearPlantillaRequest(){
		//Creamos la plantilla de respuesta a REQUEST
		MessageTemplate protocolo = MessageTemplate.MatchProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
        MessageTemplate performativa = MessageTemplate.MatchPerformative(ACLMessage.REQUEST);
        MessageTemplate plantilla = MessageTemplate.and(protocolo, performativa);
        
        return plantilla;
    
	}
	/**
	 * <p>Elimina los Elementos X e Y</p>
	 * <p>Elimina todos los elementos X e Y</p>
	 * <p>Cuando ha terminado reinicia listaControladoresAgentes,listaParejasX,listaParejasY,agentesEncontrados y contadorEncontradoLimite</p>
	 */
	private class BorrarAgentes extends Behaviour{
		
		private static final long serialVersionUID = -70230233176366781L;
		boolean finalizado=false;		
		int i=0;
		
		@Override
		public void action() {
			AgentController controlador=(AgentController) listaControladoresAgentes.get(i);
			try {
				controlador.kill();
			} catch (StaleProxyException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			i++;
			finalizado=i>=listaControladoresAgentes.size();
		}
		

		@Override
		public boolean done() {
			if (finalizado){
				listaControladoresAgentes=new ArrayList();
				listaParejasX=new ListaParejas();
				listaParejasY=new ListaParejas();
				agentesEncontrados=0;
				contadorEncontradoLimite=0;
				System.out.println("Terminada la destrucción de agentes");
			}
			return finalizado;
		}
		
	}
	
	/**
	 * <p>Obtiene los datos introducidos por el usuario y los asigna a las variables:</p>
	 * <ul>
	 * 	<li>numeroParejas </li>
	 *  <li>numeroCaracteristicas</li>
	 *  <li>filePruebas</li>
	 *  <li>numeroIteraciones</li>
	 *  <li>GaleShapleyCompleto</li>
	 *  <li>newfile</li>
	 * </ul>
	 * @see #getAllUserData(Object[] argsIn)
	 */
	private void getUserData(){
		

		//Obtenemos los argumentos por consola (opcionales)
		Object[] argumentosPorConsola = getArguments();
		
		//Gestionamos esos resultados y los asignamos a las variables
		getAllUserData(argumentosPorConsola);
		
			 
		  
	}
	/**
	 * <p>Manejador para responder Requests</p>
	 * <p>Si el contenido es "shutdown" cierra la plataforma</p>
	 *<p> Si el contenido es continueProgram (enviado por un emparejador):
	 *<ul>
	 *<li>Si no está en la última iteración, responde un Refuse. Si además se ha recibido el request de todos los emparejadores borra la generación anterior de X e Y y crea una nueva</li>
	 *<li>Si está en la última iteración responde con un agree y cuando ha recibido el request de todos los emparejadores borra la generación de X e Y</li>
	 *</ul>
	 *</p>
	 *
	 * @see Funciones#cerrarSistema
	 * @see BorrarAgentes
	 * @see crearElementosXY
	 * @see BusquedaDeElementos
	 * @see sendListasParejas
	 */
	class ManejadorResponderRequests extends AchieveREResponder
    {
       	private static final long serialVersionUID = 6898700980536834390L;
       	private int nMensajesContinue=0;

		public ManejadorResponderRequests(Agent a,MessageTemplate mt) {
            super(a,mt);
        }
 
        protected ACLMessage handleRequest(ACLMessage request)throws NotUnderstoodException, RefuseException
        {
         System.out.println("hemos recibido un mensajeRequest de "+request.getSender().getLocalName());
            String contenido=request.getContent();
            
            switch (contenido){			   
					
			case "shutdown"://Se recibe request the shutdown
				
					//Si el sistema no estaba ya cerrándose lo hace
					if (!shutdown){
						shutdown=true; //Actualizamos el valor para que no se ejecute con más llamadas el cierre del sistema
						ACLMessage agree = request.createReply();
	                    agree.setPerformative(ACLMessage.AGREE);
	                    System.out.println("Cerramos el sistema");
						Funciones.cerrarSistema(myAgent);
	                    return agree;               
					
	                    //Si ya estaba cerrándose se genera una RefuseExcepcion
					}else 
					{
						 throw new RefuseException("Sistema ya se está cerrando");
					}
					
			case "continueProgram"://Se recibe continuar el programa
				
				//Aumentamos el número de mensajes continue
				nMensajesContinue++;
				
				//Si estamos en la última iteración enviamos un Agree
				if (iteracionActual==numeroIteraciones){ 
					
					//Si además se han completado todos los emparejadores se borran los agentes
					if (nMensajesContinue==numeroEmparejadores){
						//Borramos los agentes
						addBehaviour(new BorrarAgentes());
						}
				
					//Mensaje Agree
					ACLMessage agree = request.createReply();
	                agree.setPerformative(ACLMessage.AGREE);
	                
					return agree;
					
				} 
				//No estamos en la última generación
				else {
					
					//Si se han recibido todos los emparejadores crea una nueva generación
					if (nMensajesContinue==numeroEmparejadores){
							iteracionActual++;
							nMensajesContinue=0;
							//Creamos los Elementos X e Y
							  contadorParejas=0;//Reiniciamos el contadorParejas
							  //Creamos un comportamiento secuencial, cada subcomportamiento se ejecutará tras completar el anterior
							   SequentialBehaviour seq = new SequentialBehaviour();
							   seq.addSubBehaviour(new BorrarAgentes());
							   seq.addSubBehaviour(new crearElementosXY());
								seq.addSubBehaviour(new BusquedaDeElementos());
								seq.addSubBehaviour(new sendListasParejas());
							  addBehaviour(seq);
					  
					}  
					  throw new RefuseException("Se inicia una nueva iteración");		
				}
				
				
			default: System.out.println("Mensaje no entendido : "+contenido);
			 throw new NotUnderstoodException("No se ha entendido el mensaje");					
						
            }
			}
        
        
        /**
         * Retorna nulo y no realiza ninguna acción porque el agente que envía el request no necesita información del resultado
         * Lo incluimos para completar la estructura del manejador
         */
        protected ACLMessage prepareResultNotification(ACLMessage request,ACLMessage response) throws FailureException{
			return null;
        	
        }
            
    }
	/**
	 * <p>Comportamiento que crea tantos elementos X e Y como indica la variable numeroParejas</p>
	 * <p>A cada uno de estos elementos les pasa como argumentos el número de parejas, el número de características que han de tener, la iteración actual, si las preferencias han de estar moduladas o no y el AID del gestor de agentes</p>
	 * <ul>
	 * 					<li>args[0]=gestor.getNumeroParejas();</li>
						<li>args[1]=gestor.getNumeroCaracteristicas();</li>					
						<li>args[2]=iteracionActual;</li>
						<li>args[3]=moduladas;</li>
						<li>args[4]=AID;</li>
	 * </ul>
	 * <p>El nombre de cada elemento será "elemento" seguido de X o Y según su clase y un número</p>
	 *
	 */
	class crearElementosXY extends Behaviour{
		
		private static final long serialVersionUID = 5275921196615443147L;
		
		@Override
		public void action() {
			
			//Creamos los AgentController, uno para el agente X y otro para el agente Y que se crearán
			ContainerController cc = getContainerController();
		       AgentController acX;
		       AgentController acY;
			try {
				
						//CREAMOS AGENTE X
				
						//Nombre que se pondrá al agente
						String nameX="elementoX_"+numeroAgenteNombre;
						
						/*Asignamos los argumentos que se pasarán createNewAgent) y a través de él a los agentes creados
						 * El número de parejas, el número de características, la iteración actual y si las preferencias han de estar moduladas o no
						 */
						Object[] args=new Object[5];
						GestorAgentes gestor=(GestorAgentes) esteAgente;
						
						args[0]=gestor.getNumeroParejas();
						args[1]=gestor.getNumeroCaracteristicas();
					
						args[2]=iteracionActual;
						
						args[3]=moduladas;
						args[4]=esteAgente.getAID();
						
						//Creamos el agente y lo iniciamos
						acX = cc.createNewAgent(nameX, "tfgMarriageStable.ElementoX", args);
						acX.start();
						
						//Añadimos el controlador a la lista de agentes
					    listaControladoresAgentes.add(acX);
						
					    //CREAMOS AGENTE Y
					    
					    String nameY="elementoY_"+numeroAgenteNombre;
						//Creamos el agente y lo iniciamos
						acY = cc.createNewAgent(nameY, "tfgMarriageStable.ElementoY", args);
						acY.start();
						
						//Añadimos el controlador a  la lista de agentes
						listaControladoresAgentes.add(acY);
						
						
						//Actualizamos los agentes creados y el número que se añadirá al nombre del agente
						contadorParejas++;
						numeroAgenteNombre++;					
						
			} 
			
			//Excepción que se produce si no se ha podido crear el agente
			catch (StaleProxyException e) {
				
				/*Si el nombre del agente ya está en la plataforma
				 * Incrementamos el número en el nombre del agente a crear para que el siguiente tenga otro nombre	
				 **/
				 
				if (e.getMessage().contains("already present in the platform")){
					numeroAgenteNombre++;	
				}
				
				System.out.println("No se ha podido crear el agente\n"+e.getMessage());
			}
			
		}

		@Override
		public boolean done() {
			if (contadorParejas==numeroParejas){
				System.out.println("Se han creado "+contadorParejas+" parejas");
			}
			return contadorParejas==numeroParejas;
		}
		
	}
	
	/**
	 * Obtiene el número de parejas
	 * @return int número de parejas
	 */
	private int getNumeroParejas(){
		return this.numeroParejas;
	}
	
	
	/**
	 * Obtiene el número de características
	 * @return int el número de características
	 */
	private int getNumeroCaracteristicas(){
		return this.numeroCaracteristicas;
	}
	
	
/**	
 * <p>Comportamiento que busca los elementos X e Y de la iteración actual</p>
 * <p>Añade los elementos encontrados a listaParejasX y listaParejasY respectivamente</p>
 * <p>Si no encuentra los elementos se borra el agente y se asigna la variable faltanParejas=true</p>
 *
 */
private class BusquedaDeElementos extends Behaviour{ 	
	
	
		private static final long serialVersionUID = 2830350408286431010L;
		boolean completado=false;//Variable que nos sirve para comprobar la finalización del comportamiento
    	@Override
    	public void action() {
    		System.out.println(myAgent.getLocalName()+" buscando parejas");
    					
    		/*Comprobamos el valor de contadorEncontradoLimite. 
    		 * Si supera el límite quiere decir que siempre obtenemos el mismo numero sin llegar al esperado,
    		 *  probablemente porque no se ha aumentado aumentado el valor por defecto de las búsquedas.
    		 * Se eliminará el agente
    		*/
    		if (contadorEncontradoLimite>NUMERO_INTENTOS_BUSQUEDAS){
    		
    			System.err.println("\n"+myAgent.getLocalName()+" no ha encontrado todas las parejas, se borrará");
    			   			
    			
    			faltanAgentes=true;
    			//Borramos el agente
    			doDelete(); 			
    			
    		}    	
    	
    		//Creamos el modelo sobre el que se realizarán las consultas
	    	DFAgentDescription template=new DFAgentDescription();
	    	
	    	//Servicio que se proporciona	    	
	    	ServiceDescription templateSd = new ServiceDescription();
	    	String tipo="parejas"+iteracionActual;
	    			    	
			templateSd.setType(tipo);
	  		template.addServices(templateSd);	
	    	
	    	try
	    	{
	    		//Creamos la consulta
	    		SearchConstraints sc = new SearchConstraints();
	    		
	    		// Seleccionamos cuántos resultados esperamos como máximo
	    		sc.setMaxResults(new Long(numeroParejas*2));
	    		   			    	    	
	    	//Hacemos la búsqueda indicando al agente los criterios de selección
	    	DFAgentDescription[] result = DFService.search(myAgent, template, sc);
	    	
	    	//Si el número de resultados no es igual al número de parejas esperadas
	    	if (result.length!=numeroParejas*2){	    		
	    		
	    		
	    		//Caso normal, se actualiza el número de agentesEncontrados
	    		if (result.length>agentesEncontrados){
	    			agentesEncontrados=result.length;
	    		}else{
	    					//Se repite el número de agentes encontradas se aumenta el contador del límite y se muestra un mensaje de error
	    			 if ((agentesEncontrados==result.length)){
	    				 contadorEncontradoLimite++; 
	    				 System.err.println("Número de agentes encontrados se ha repetido en la siguiente búsqueda. Agente: "+myAgent.getLocalName()+" se esperaban "+numeroParejas*2+" y se han encontrado "+result.length+" Quedan "+(NUMERO_INTENTOS_BUSQUEDAS-contadorEncontradoLimite+1)+" intentos");
	    			 }
	    			 
	    			 //Si se encuentran menos parejas que en un intento anterior se borra el agente
	    			 if ((agentesEncontrados>result.length)){
	    				 System.err.println("Ha habido algún problema en la búsqueda de parejas "+"El agente "+myAgent.getLocalName()+ "se cerrará\nCompruebe que no ha creado más parejas de las que se pueden buscar (por defecto 100)\nSi desea modificar ese valor hágalo al ejecutar la plataforma JADE añadiendo el parámetro -jade_domain_df_maxresult NUEVOVALOR, ejemplo:-jade_domain_df_maxresult 700");
	    				 faltanAgentes=true;
	    				 doDelete();
	    				
	    			 }
	    		}
	    	}
	    	//Si coinciden, como se espera se añaden a la lista de agentes conocidos por el agente
	    	else{
	    	System.out.println("Se han encontrado "+result.length+" elementos");
	    	for (int i = 0; i < result.length; ++i) {
	  				DFAgentDescription dfds = result[i];
	  			//Obtenemos el AID del agente encontrado	
				AID providerAID = dfds.getName();
				//Creamos un ElementoValorado con ese nombre y una valoración inicial de 0
				ElementoValorado provider=new ElementoValorado(providerAID,(float)0.0);
				
				
				/*El mismo agente puede proporcionar varios servicios solo estamos interesados
				en el servicio "parejas"*/					
				@SuppressWarnings("unchecked")
				Iterator<ServiceDescription> it = dfds.getAllServices();
				
				
				while (it.hasNext()) {
					
					ServiceDescription sds = it.next();
					 tipo="parejas"+iteracionActual;
			    	
					if (sds.getType().equals(tipo)) {						
						
						if (sds.getName().equals("conocer_parejasY")){
							
							listaParejasY.add(provider);
							
						}
						//Añadimos a la lista de agentes encontrados
						if (sds.getName().equals("conocer_parejasX")){
							
							listaParejasX.add(provider);
							
						}
					
					
						
					}
					
				}
				
				
				completado=true;//Hacemos que el comportamiento pueda finalizar
	    	}
	    	}
	    	}catch(FIPAException e){
	    		System.err.println("Error temporal en la búsqueda de agentes en el DF del agente: "+myAgent.getLocalName());
	    		System.err.println(e.getMessage());
	    	}
    		}
    	@Override
    	public boolean done() { 
    		
    		return completado;//Retorna la variable completado, si su valor es false el comportamiento se volverá a ejecutar
    		
    	}
}

/**
 * Comportamiento que envía un mensaje inform a cada miembro de las listas listaParejasY
 	y istaParejasX con el conversationId "listaParejas"
 *  *
 */
@SuppressWarnings("serial")
private class sendListasParejas extends Behaviour{

private boolean completo=false;
	@Override
	public void action() {

		//Creamos el mensaje
	   ACLMessage msgx = new ACLMessage(ACLMessage.INFORM);
		 //Añadimos el lenguaje empleado
		msgx.setLanguage(new SLCodec().getName());
		ACLMessage msgy = new ACLMessage(ACLMessage.INFORM);
			 //Añadimos el lenguaje empleado
			msgx.setLanguage(new SLCodec().getName());
			 //Añadimos el lenguaje empleado
			msgy.setLanguage(new SLCodec().getName());
			msgx.setConversationId("listaParejas");
			msgy.setConversationId("listaParejas");
		try {
			msgx.setContentObject(listaParejasY.toAID_List());
			msgy.setContentObject(listaParejasX.toAID_List());
			
			int i=0;
			while (i<listaParejasY.size()){
				
				AID receptorX = (AID) listaParejasX.toAID_List().get(i);
				AID receptorY = (AID) listaParejasY.toAID_List().get(i);
				System.out.println("Añadiendo receptores "+receptorX.getLocalName()+","+receptorY.getLocalName());
				msgx.addReceiver(receptorX);
				msgy.addReceiver(receptorY);
				i++;
			}
			
			
			send(msgx);
			send(msgy);
			
			completo=true;;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			System.err.println("No se ha podido enviar la lista de parejas");
			e.printStackTrace();
			completo=true;
		}
		
	}
	@Override
	public boolean done() {
		// TODO Auto-generated method stub
		return completo;
	}
}
/**
 * Abre un cuadro de diálogo con el usuario preguntando cuántas parejas quiere crear
 * <p>Si son más de MAXIMO_RESULTADO_BUSQUEDAS_DEFECTO(el límite por defecto de las búsquedas es 100, instruye al usuario cómo cambiarlo</p>
 *<p> El usuario tiene la opción de cerrar la plataforma para reiniciarla con un nuevo máximo.</p>
 * @return int  el número de parejas introducido por el usuario
 * @see Funciones#cerrarSistema
 */
private  int getNumeroParejasFromUser(){
	
	//Mientras el número no sea estrictamente positivo y como máximo PAREJAS_MAXIMAS_PERMITIDAS  se seguirá pidiendo
	while (numeroParejas<=0||numeroParejas>PAREJAS_MAXIMAS_PERMITIDAS ){
		try{
			
			//Obtenemos el número del usuario
			numeroParejas =  Integer.parseInt(JOptionPane.showInputDialog("¿Cuántas parejas quiere crear?"));
			
			//Número no puede ser negativo
			if (numeroParejas<0){
				JOptionPane.showMessageDialog(null,"El número de parejas no puede ser negativo \nInténtelo de nuevo");
						
		}
			//Sobrepasamos máximo de 
			if (numeroParejas>MAXIMO_RESULTADO_BUSQUEDAS_DEFECTO/2){
				
				int i =JOptionPane.showConfirmDialog(null,"\nHa seleccionado un número de parejas superior al valor que por defecto obtendrán los resultados de las búsquedas("+MAXIMO_RESULTADO_BUSQUEDAS_DEFECTO/2+")\nPara que el programa funcione ha de haber configurado la plataforma aumentando ese número al ejecutarla\n¿Está seguro?","Confirmar número superior a "+MAXIMO_RESULTADO_BUSQUEDAS_DEFECTO/2,JOptionPane.YES_NO_OPTION);
				if (i==0){
					//No hacemos nada
				}
				if (i==1){
					numeroParejas=0;
					int j =JOptionPane.showConfirmDialog(null,"\n¿Desea información sobre cómo aumentar el número posible de búsquedas hasta "+PAREJAS_MAXIMAS_PERMITIDAS+"?","¿Ejemplo?",JOptionPane.YES_NO_OPTION);
					
						if (j==0){
							JOptionPane.showMessageDialog(null,"Al ejecutar la plataforma ha de añadir el parámetro\n -jade_domain_df_maxresult NUMERO_PAREJAS*2 (RECUERDE: SE HA DE BUSCAR EL DOBLE DE ELEMENTOS QUE DE PAREJAS)\n");
							String[] options = {"Introducir un nuevo número", "Cerrar la plataforma"};
							int seleccion = JOptionPane.showOptionDialog(null, "¿Qué desea hacer?", "Opciones", JOptionPane.DEFAULT_OPTION, JOptionPane.QUESTION_MESSAGE, null, options, options[0]);
							if (seleccion==1){
								Funciones.cerrarSistema(esteAgente);
								
							}else numeroParejas=0;
								
						}else
						{JOptionPane.showMessageDialog(null,"Vuelva a intentarlo, por favor");
						}
						
				}
				
			}
			if (numeroParejas>PAREJAS_MAXIMAS_PERMITIDAS ){
				JOptionPane.showMessageDialog(null,"El número de parejas no puede ser mayor de "+PAREJAS_MAXIMAS_PERMITIDAS+"\nInténtelo de nuevo");
						
		}
		}catch (Exception e){
			
			if (e.getMessage()!="null"){				
			
			JOptionPane.showMessageDialog(null,"Ha habido algún error\nInténtelo de nuevo");
			numeroParejas=0;			
			}else {		
					numeroParejas=-1;
					JOptionPane.showMessageDialog(null,"Ha cancelado la creación de parejas\nEl gestor de agentes se cerrará");
					esteAgente.doDelete();	
					break;				
					
				}
		}
	}
	return numeroParejas;//El número de parejas introducido
}	 
/**
 * Abre un cuadro de diálogo para obtener del usuario el número de características del agente
 * <p>Si se cancela la entrada de características cierra el gestor de agentes</p>
 * @return int, el número de características introducido por el usuario (entre 1 y NUMERO_MAXIMO_CARACTERISTICAS)
 */
private int getNumeroCaracteristicasFromUser(){
	while (numeroCaracteristicas<=0||numeroCaracteristicas>NUMERO_MAXIMO_CARACTERISTICAS ){
		try{
			
			//Obtenemos el número del usuario
			numeroCaracteristicas =  Integer.parseInt(JOptionPane.showInputDialog("¿Cuántas características quiere crear?"));
			
			if (numeroCaracteristicas<0){
				JOptionPane.showMessageDialog(null,"El número de características no puede ser negativo \nInténtelo de nuevo");
					
		}
		    //Sobrepasamos máximo
					if (numeroCaracteristicas>NUMERO_MAXIMO_CARACTERISTICAS){
						JOptionPane.showMessageDialog(null,"El número de características no puede ser mayor de "+NUMERO_MAXIMO_CARACTERISTICAS+" \nInténtelo de nuevo");
						
					}
			
			
		}
		catch (Exception e){						
			if (e.getMessage()!="null"){				
			
			JOptionPane.showMessageDialog(null,"Ha habido algún error\nInténtelo de nuevo");
			numeroCaracteristicas=0; 
			}else{
				numeroCaracteristicas=-1;
				JOptionPane.showMessageDialog(null,"Ha anulado la creación de parejas\n el gestor de agentes se cerrará");
				esteAgente.doDelete();	
				break;
			}
		}
	}
		return numeroCaracteristicas;
}
/**
 * <p>Gestiona todos los datos introducidos por el usuario</p>
 * <p>Obtiene el número de parejas y el número de características</p>
 * <p>Gestiona los datos introducidos como argsIn
 * <ul>
	 * <li>Si existe en argsIn un argumento con valor "ft" asigna al campo <i>filePruebas</i> el valor true</Li>
	 * <li>Si existe en argsIn un argumento que sea un entero asigna a la variable <i>numeroIteraciones</i> el valor de ese entero</Li>
	 * <li>Si existe en argsIn un argumento con valor "completo" asigna a la variable <i>GaleShapleyCompleto</i> el valor true;</Li>
	 * <li>Si existe en argsIn un argumento con valor "moduladas" asigna a la variable <i>moduladas</i> el valor true;</li>
	 * 
 * </ul>
 * 
 * @param argsIn Object[]  Se esperan argumentos introducidos por consola al crear un gestor de agentes
 * @see #getNumeroParejasFromUser()
 * @see #getNumeroCaracteristicasFromUser()
 */
private void getAllUserData(Object[] argsIn){
	

	//Si se han recibido argumentos
	if (argsIn!=null) {	
									  
		//Inicialmente ajustamos el número de iteraciones a 1
		int iteraciones=1;
		
		//Recorremos los argumentos recibidos
		int i=0;
		while (i<argsIn.length){
			
			//Convertimos el argumento obtenido en un String
			String valor = (String) argsIn[i];
			
			//Comprobamos si es un número, si lo es, se asigna su valor a numeroIteraciones
			//Además se guarda como valor="iteraciones"
			try {
				iteraciones=Integer.parseInt(valor);
				valor="iteraciones";
			}catch (Exception e){
				//No hacemos nada
			}
			
			//Comprobamos los diferentes casos
			switch (valor) {
			
					//Se guardarán los archivos de muestras de preferencias
					case "fp":
						filePruebas=true;
						System.out.println("Se guardarán los archivos con preferencias para pruebas");
					break;
					
					//Se guardarán el número de iteraciones
					case "iteraciones":
						this.numeroIteraciones=iteraciones;
					break;
					
					//Se envía un true (para crear un GaleShapleyCompleto)	
					case "completo":
						GaleShapleyCompleto=true;
					break;
					
					//Se envía un true (para guardar un nuevo archivo)
					case "newfile":
						newfile=true;
					break;
					
					//Se actualiza el valor de moduladas
					case "moduladas":
						moduladas=true;
					break;
					
			}
			i++;
						
		}
			
	}
	
	//Obtenemos el número de parejas
	int parejas= getNumeroParejasFromUser();
	
	//Si es un valor válido obtenemos el número de características
	if (parejas>0){
		this.numeroParejas=parejas;
	int caracteristicas=getNumeroCaracteristicasFromUser();
	this.numeroCaracteristicas=caracteristicas;
	}else {
		this.numeroCaracteristicas=0;
	}
	
}

/**
 * <p>Comportamiento que escucha un conversationId que se marca al iniciarlo</p>
 * <p>Si el contenido es un long, la primera vez actualiza el valor de <i>time_start</i> con ese long</p>
 * <p>Las siguientes veces incrementa en 1 el campo <i>mensajesElementos</i> hasta que ha recibido el núemero de parejas por 2</p>
 * <p>En ese momento actualiza con el nuevo long el campo time_end y escribe en el archivo comunicaciones.csv el número de parejas y la diferencia entre ambos tiempos</p
 * <p>Por último, finaliza el comportamiento, que solo se producirá en la primera iteración porque las siguientes los tiempos quedan desvirtuados por el uso de más recursos</p>
 *
 */
private class EscucharConversationId extends Behaviour{
	/**
	 * 
	 */
	private static final long serialVersionUID = 6599371313619065253L;

	/**
	 * Identificador de conversación en el que escucha el comportamiento
	 */
	private String conversationId;
	
	/**
	 * Variable que guarda el momento en el que se recibe el  mensaje del último elemento diciendo que ha completado su lista
	 */
	private long time_end;

	
	/**
	 * Variable que guarda el momento en el que se recibe el primer mensaje de un elemento diciendo que ha completado su lista
	 */
	private long time_start=0;
	
	/**
	 * Guarda el número de mensajes de los elementos recibidos
	 */
	private int mensajesElementos=0;
	
	
	private boolean completado=false;
	
	private EscucharConversationId(String pConversation){
		this.conversationId=pConversation;
	}
	@Override
	public void action() {
		//Creamos un filtro, solo se recibirán los mensajes cuyo identificador de conversacion coincida con conversationId
		
				MessageTemplate mt=MessageTemplate.MatchConversationId(conversationId);
				//Recepción no bloqueante de un mensaje según el filtro creado mt
				ACLMessage msg = myAgent.receive(mt);	
		
		if (msg != null) {
			
				try {
					long time=(Long) msg.getContentObject();
				
					mensajesElementos++;
					
					//Si time_start es 0 es el primer mensaje recibido cambiamos el tiempo de inicio
					if (time_start==0){
						time_start=time;
						
					}
					
					//Si se recibe el último mensaje escribimos el archivo
					if(mensajesElementos==numeroParejas*2){
					
						time_end = System.currentTimeMillis();
					
						
						String texto=numeroParejas+","+Long.toString(time_end-time_start)+"\r\n";
						Funciones.createFile(texto, "comunicaciones.csv", true);
						completado=true;
					}
					
				
					
				} catch (UnreadableException e) {
					System.err.println("No se ha podido leer el documento");
					e.printStackTrace();
				}
			
			
						
		}else
			{
			//Bloqueamos el comportamiento hasta recibir un mensaje en la cola
			block();
			}
	  }
	
	public boolean done() {
		// TODO Auto-generated method stub
		return completado;
	}
		
	}

	
	
}


	





	

